import axios from 'axios';
// trailing slash is deliberate, to differentiate this package from the core node module
import { Buffer } from 'buffer/';
import { recoverPersonalSignature } from '@metamask/eth-sig-util';
import objectHash from 'object-hash/dist/object_hash';

// Make Buffer available to recoverPersonalSignature
window.Buffer = Buffer;

class Post {
  constructor({
    author, content, signature, hash,
  }) {
    this.author = author;
    this.content = content;
    this.signature = signature;
    this.hash = hash;
  }

  // Read from API
  static async read(hash) {
    const { data: { content, author, signature } } = await axios.get(`/api/read/${hash}`);
    // Verify hash
    const derivedHash = objectHash({ author, content, signature });
    if (hash !== derivedHash) {
      throw new Error('Hash mismatch');
    }
    // Verify signature
    const recovered = recoverPersonalSignature({ data: content, signature });

    if (recovered !== author) {
      throw new Error('Author mismatch');
    }
    return new Post({
      content, author, signature, hash,
    });
  }

  // Include MetaMask signature
  async sign(web3Provider, account) {
    this.author = account;
    const msg = `0x${Buffer.from(this.content, 'utf8').toString('hex')}`;
    this.signature = await web3Provider.request({
      method: 'personal_sign',
      params: [msg, account],
    });
    return this;
  }

  // Write to API
  async write() {
    const data = {
      author: this.author,
      content: this.content,
      signature: this.signature,
    };
    const { data: hash } = await axios.post('/api/write', data);
    this.hash = hash;
  }

  // Upload hash to blockchain
  async publish(DAO, account) {
    await DAO.methods.addPost(account, this.hash).send({
      from: account,
      gas: 1000000,
    });
  }
}

export default Post;
